

#! /usr/bin/env python

import os
import sys
import os
import itertools
import argparse
import errno
import math
import copy


from collections import namedtuple


from modules import range_query_max_search_tree as RMaxQST


def argmax(iterable):
    return max(enumerate(iterable), key=lambda x: x[1])[0]

def max_both(iterable):
    return max(enumerate(iterable), key=lambda x: x[1])

def traceback(index, C):
    return 

def all_solutions_c_max_indicies(C, C_max):
    return [i for i, c in enumerate(C) if c == C_max] 

# def all_solutions_c_max_indicies_mam(C, C_max):
#     return [i for i, c in enumerate(C) if c >= C_max - 1 ] 

# def reconstruct_solution(mems, C, trace_vector):
#     solution_index = argmax(C)
#     value = C[solution_index]
#     # print()
#     solution = []
#     while solution_index > 0:
#         solution.append(mems[solution_index - 1])  # trace vector is shifted on to the right so need to remove 1 from vectore to get j_index 
#         solution_index = trace_vector[solution_index]
#         # print(solution_index)
#         # if solution_index is None:
#         #     break
#     return value, solution[::-1]

def reconstruct_all_solutions(mems, all_C_max_indicies, trace_vector, C, mam_mode = False):
    # solution_index = argmax(C)
    solutions = []
    for solution_index in all_C_max_indicies:
        value = C[solution_index]
        solution = []
        while solution_index > 0:
            if mam_mode:
                solution.append(mems[solution_index])  # trace vector is shifted on to the right so need to remove 1 from vectore to get j_index 
            else:
                solution.append(mems[solution_index - 1])  # trace vector is shifted on to the right so need to remove 1 from vectore to get j_index 
            solution_index = trace_vector[solution_index]
        solutions.append( solution[::-1] )
    return value, solutions

def make_leafs_power_of_2(mems):
    nodes = []
    nodes.append( RMaxQST.Node(-1, -1, -2**32, -1) ) # add an start node in case a search is smaller than any d coord in tree
    for i, mem in enumerate(mems):
        # if i > 3: break
        m = RMaxQST.Node(mem.d, mem.j, -2**32, mem.j)
        nodes.append(m)

    for i in range(20):
        if len(nodes) == 2**i or len(nodes) == 2**(i+1):
            break
        elif 2**i < len(nodes) < 2**(i+1):
            remainder = 2**(i+1) - len(nodes) 
            for i in range(remainder):
                nodes.append( RMaxQST.Node(-1, -i - 2, -2**32, -i - 2) ) # fill up nodes to have leaves a power of 2
            break

    leafs = sorted(nodes, key= lambda x: x.d)
    # n = len(leafs)
    return leafs

def n_logn_read_coverage(mems):
    """
        Algorithm 15.1 in Genome scale algorithmic design, Makinen et al.

        Using Binary search trees for range max queries,
        so n log n time complexity. Each mem is an Namedtuple. python object

    """
    # assert mems == sorted(mems, key=lambda x: x.y)


    T_leafs = make_leafs_power_of_2(mems)
    I_leafs = make_leafs_power_of_2(mems)
    n = len(T_leafs)
    T = [0 for i in range(2 * n) ]  
    I = [0 for i in range(2 * n) ]  
    # T_leafs = copy.deepcopy(leafs)
    RMaxQST.construct_tree(T, T_leafs, n)
    # I_leafs = copy.deepcopy(leafs)
    RMaxQST.construct_tree(I, I_leafs, n)

    mem_to_leaf_index = {l.j : i for i,l in enumerate(T_leafs)}

    C = [0]* (len(mems) + 1) #(len(leafs))
    trace_vector = [None]*(len(mems) + 1)

    RMaxQST.update(T, 0, 0, n) # point update 
    RMaxQST.update(I, 0, 0, n) # point update 

    for j, mem in enumerate(mems):
        leaf_to_update = mem_to_leaf_index[j]
        c = mem.c
        T_max, j_prime_a, node_pos  = RMaxQST.range_query(T, -1, c-1, len(T_leafs)) 
        # print("C_a:",  T_max +  mem.d - mem.c + 1, j_prime_a, node_pos, leaf_to_update )
        # print("T TREE:", [(s, zz.j, zz.d, zz.Cj, zz.j_max) for s, zz in enumerate(T) if type(zz) != int])
        C_a =  T_max +  mem.d - mem.c + 1  # add the mem_length to T since disjoint

        if T_max < 0:
            print("BUG", T_max)
            sys.exit()

        
        d = mem.d
        I_max, j_prime_b, node_pos  = RMaxQST.range_query(I, c, d, len(I_leafs))         
        # print("C_b:", I_max +  mem.d, I_max, j_prime_b, node_pos, leaf_to_update )
        # print( I_max, mem.d, mems[j_prime_b].d, mems[j_prime_b])
        # print("I TREE:", [(s, zz.j, zz.d, zz.Cj, zz.j_max) for s, zz in enumerate(I) if type(zz) != int])
        C_b =  I_max +  mem.d #- mems[j_prime_b].d   # add the part of the mem that is not overlapping

        # if C_b < 0:
        #     print("BUG")
        #     sys.exit()

        index, value = max_both([C_a, C_b])
        C[j+1] = value
        if index == 0: # Updating with C_a
            j_prime = j_prime_a
        else: # Updating with C_b
            j_prime = j_prime_b


        if j_prime < 0: # any of the additional leaf nodes (with negative index) we add to make number of leafs 2^n
            trace_vector[j+1] = 0
        elif value == 0: # first j (i.e. j=0) 
            trace_vector[j+1]= 0
        else:
            trace_vector[j+1] = j_prime +1

        RMaxQST.update(T, leaf_to_update, value, n) # point update 
        RMaxQST.update(I, leaf_to_update, value - mem.d, n) # point update 


    # C_max, solution = reconstruct_solution(mems, C, trace_vector)
    # print("C", C)
    # print(trace_vector)

    solution_index = argmax(C)
    C_max = C[solution_index]
    all_C_max_indicies = all_solutions_c_max_indicies(C, C_max)
    # print("number solutions with the same score:", all_solutions_c_max_indicies(C, C_max))
    C_max, solutions = reconstruct_all_solutions(mems, all_C_max_indicies, trace_vector, C)

    return solutions, C_max #, is_unique_solution(C)

def read_coverage(mems, max_intron):
    """
        Algorithm 15.1 in Genome scale algorithmic design, Makinen et al.

        Using lists instead of Binary search trees for range max queries,
        so n^2 time complexity instead of O(n log n).

        each mem is an Namedtuple. python object

    """
    # assert mems == sorted(mems, key = lambda x: x.y )

    if len(mems) > 1000:
        print('MEM',len(mems))

    # print("Going in to mem chaining:", mems)
    T = [ (v.d, v.val)  for v in mems]
    I = [ (v.d, v.val)  for v in mems]
    
    # T_dict = {mems[i][3] : -10000 for i in range(len(mems))}
    # T_dict[0] = -10000
    # I_dict = {mems[i][3] : -10000 for i in range(len(mems))}
    # I_dict[0] = -10000

    # C_a = [0]*(len(T))
    # C_b = [0]*(len(T))
    C = [0]*(len(T)+1)
    traceback_vector = [None]*(len(T)+1)

    for j in range(len(T)):
        v =  mems[j]

        # linear scan -- replace with range max Q tree
        T_values = [(j_prime, c_val) for j_prime, c_val in enumerate(C[1:]) if  mems[j_prime].d < v.c and j_prime < j and v.y - mems[j_prime].y < max_intron ]
        if T_values:
            # print(T_values)
            T_traceback_index, max_c_value_case_a = max(reversed(T_values), key=lambda x: x[1])
        else:
            max_c_value_case_a = 0
            T_traceback_index = -1

        I_values = [(j_prime, c_val) for j_prime, c_val in enumerate(C[1:]) if v.c <= mems[j_prime].d  <= v.d and j_prime < j and v.y - mems[j_prime].y < max_intron ]
        if I_values:
            # print(I_values)
            I_values_plus_chord_diff = [ (j_prime, c_val + (v.d - mems[j_prime].d)) for j_prime, c_val in I_values]
            I_traceback_index, max_c_value_case_b = max(reversed(I_values_plus_chord_diff), key=lambda x: x[1])
            # I_v_prev_coord = mems[I_traceback_index].d
            # C_b[j] = (v.d - I_v_prev_coord) + max_c_value_case_b # shouldnt it be v.d - v_tmp.d
            C_b = max_c_value_case_b # shouldnt it be v.d - v_tmp.d

        else:
            I_v_prev_coord = v.c - 1
            I_traceback_index = -1
            max_c_value_case_b = 0
            C_b = 0


        C_a = (v.d - v.c + 1) +  max_c_value_case_a

        index, value = max_both([C_a, C_b])
        C[j+1] = value
        if index == 0: # Updating with C_a
            j_prime = T_traceback_index
        else: # Updating with C_b
            j_prime = I_traceback_index

        # C[j+1] = max(C_a, C_b)

        if j_prime < 0: # first j (i.e. j=0) 
            traceback_vector[j+1]= 0
        else:
            traceback_vector[j+1]= j_prime + 1

        # if j_prime == 0: # first j (i.e. j=0) 
        #     traceback_vector[j+1]= 0
        # elif C_a >= C_b:
        #     traceback_vector[j+1] = T_traceback_index + 1
        # else:
        #     traceback_vector[j+1] = I_traceback_index + 1

        # print(v.c, v.d, v.d -v.c, C_a[j], C_b[j], v.d, I_values, T_values)


    # solution_index = argmax(C)
    # value = C[solution_index]
    # solution = []
    # while True:
    #     solution.append(mems[solution_index])
    #     solution_index = traceback_vector[solution_index]
    #     if solution_index is None:
    #         break


    solution_index = argmax(C)
    # print(C)
    # print(traceback_vector)
    C_max = C[solution_index]
    all_C_max_indicies = all_solutions_c_max_indicies(C, C_max)
    # print(all_C_max_indicies)
    # print("number solutions with the same score:", all_solutions_c_max_indicies(C, C_max))
    C_max, solutions = reconstruct_all_solutions(mems, all_C_max_indicies, traceback_vector, C)
    # solutions = []
    # print("MEM Solution:", solution[::-1])
    return solutions, C_max #, is_unique_solution(C)
    # traceback(C, best_solution_index)


def read_coverage_mam_score(mams, overlap_threshold = 20):
    """
        Algorithm 15.1 in Genome scale algorithmic design, Makinen et al.

        Using lists instead of Binary search trees for range max queries,
        so n^2 time complexity instead of O(n log n).

        each mem is an Namedtuple. python object

    """

    # assert mams == sorted(mams, key=lambda x: x.y)
    # mams = sorted(mams, key = lambda x: x.y )

    # print("MAM INSTANCE", mams)
    # for mam in mams:
    #     print(mam.mam_id, mam.x, mam.y, mam.c, mam.d, '\t', mam.val, mam.min_segment_length)
    if len(mams) > 1000:
        print('MAM',len(mams))
    T = [ (v.d, v.val)  for v in mams]
    I = [ (v.d, v.val)  for v in mams]
    
    # T_dict = {mams[i][3] : -10000 for i in range(len(mams))}
    # T_dict[0] = -10000
    # I_dict = {mams[i][3] : -10000 for i in range(len(mams))}
    # I_dict[0] = -10000

    C_a = [0]*(len(T))
    C_b = [0]*(len(T))
    C = [0]*(len(T))
    traceback_pointers = [None]*(len(T))

    for j in range(len(T)):
        v =  mams[j]

        # linear scan -- replace with range max Q tree
        T_values = [(j_prime, c_val - 0.1* (v.c - mams[j_prime].d - 1) ) for j_prime, c_val in enumerate(C) if  mams[j_prime].d < v.c and j_prime < j]
        # T_values2 = [(j_prime, c_val - max(0, mams[j_prime].y - v.x)) for j_prime, c_val in enumerate(C) if  mams[j_prime].d < v.c and j_prime < j] # Is this proper symmetric variant subtracting overlapping genomic positions?
        if T_values:
            # print(j, T_values)
            T_traceback_index, max_c_value_case_a = max(reversed(T_values), key=lambda x: x[1])
        else:
            max_c_value_case_a = 0
            T_traceback_index = None

        # I_values = [(j_prime, c_val) for j_prime, c_val in enumerate(C) if v.c <= mams[j_prime].d <= v.d and j_prime < j]
        I_values = [(j_prime, c_val) for j_prime, c_val in enumerate(C) if v.c <= mams[j_prime].d <= v.c - 1 + overlap_threshold and j_prime < j]
        # I_values = [(j_prime, c_val) for j_prime, c_val in enumerate(C) if v.c <= mams[j_prime].d  <= v.c + (v.d - v.c)*(1.0 - v.val/(v.d - v.c)) ]  #Maybe bug here, check the objective!! need to be careful
        if I_values:
            # print("here", j, I_values)
            # I_values_plus_chord_diff = [ (j_prime, c_val + (v.val - (mams[j_prime].d - v.c + 1 ) - 0.01 if v.x != mams[j_prime].y else v.val - 0.01)) for j_prime, c_val in I_values]  # Penalize an extra 0.1 to prefer non-overlapping solutions in case of tie breakers
            I_values_plus_chord_diff = [ (j_prime, c_val + (v.val - (mams[j_prime].d - v.c + 1 ) - 0.1*(mams[j_prime].d - v.c + 1 )) ) for j_prime, c_val in I_values]  # Penalize an extra 0.1 to prefer non-overlapping solutions in case of tie breakers

            I_traceback_index, max_c_value_case_b = max(reversed(I_values_plus_chord_diff), key=lambda x: x[1])
            C_b[j] = max_c_value_case_b

            # I_traceback_index, max_c_value_case_b = max(I_values, key=lambda x: x[1])
            # I_v_prev_coord = mams[I_traceback_index].d
            # C_b[j] = max_c_value_case_b + (v.val - v.val*(v.d - I_v_prev_coord)) # (v.d - I_v_prev_coord) +

        else:
            I_v_prev_coord = None
            I_traceback_index = None
            C_b[j] = 0

        C_a[j] = max_c_value_case_a + v.val # + (v.d - v.c + 1)

        if C_a[j] >= C_b[j]:
            traceback_pointers[j] = T_traceback_index
        else:
            # print("here")
            traceback_pointers[j] = I_traceback_index

        C[j] =  max(C_a[j], C_b[j]) #C_a[j] 

    # print('Value vector Max approx matches:', C)
    # print(traceback_pointers)

    # print(argmax(C))
    # print(traceback_pointers)
    solution_index = argmax(C)
    # solution_index = len(C) - argmax(C[::-1]) -1
    # print()
    # help_functions.eprint(C)
    value = C[solution_index]
    # print("index best sol:", solution_index, argmax(C), len(C), C)
    # print(traceback_pointers)

    # all_C_max_indicies = all_solutions_c_max_indicies_mam(C, value)
    # print(C)
    # print([m.x for m in mams])
    # print(traceback_pointers)
    # print(all_C_max_indicies)
    # print("number solutions with the same score:", all_solutions_c_max_indicies_mam(C, value))
    # C_max, solutions = reconstruct_all_solutions(mams, all_C_max_indicies, traceback_pointers, C, mam_mode = True)
    # for s in solutions:
    #     print(C_max, s)

    solution = []
    while True:
        solution.append(mams[solution_index])
        solution_index = traceback_pointers[solution_index]
        # print(solution_index)
        if solution_index is None:
            break

    non_unique = [x for x in set(C) if C.count(x) > 1]
    unique = True
    if non_unique:
        unique = False
        # print(traceback_pointers)

    # print("Solution:", solution[::-1])
    return tuple(solution[::-1]), value, unique
    # traceback(C, best_solution_index)



def n_logn_read_coverage_mams(mams, overlap_threshold = 5):
    """
        Algorithm 15.1 in Genome scale algorithmic design, Makinen et al.

        Using Binary search trees for range max queries,
        so n log n time complexity. Each mem is an Namedtuple. python object

    """
    # assert mams == sorted(mams, key=lambda x: x.y)
    # for mam in mams:
    #     print(mam.mam_id, mam.x, mam.y, mam.c, mam.d, '\t', mam.val, mam.min_segment_length)
    # overlap_threshold = 20
    T_leafs = make_leafs_power_of_2(mams)
    I_leafs = make_leafs_power_of_2(mams)
    n = len(T_leafs)
    T = [0 for i in range(2 * n) ]  
    I = [0 for i in range(2 * n) ]  
    # T_leafs = copy.deepcopy(leafs)
    RMaxQST.construct_tree(T, T_leafs, n)
    # I_leafs = copy.deepcopy(leafs)
    RMaxQST.construct_tree(I, I_leafs, n)

    mem_to_leaf_index = {l.j : i for i,l in enumerate(T_leafs)}

    C = [0]* (len(mams) + 1) #(len(leafs))
    trace_vector = [None]*(len(mams) + 1)

    # aux_score = [0]* (len(mams) + 1) #(len(leafs))

    RMaxQST.update(T, 0, 0, n) # point update 
    RMaxQST.update(I, 0, 0, n) # point update 
    for j, mam in enumerate(mams):
        # print()
        # print( 'starting:', mam.c, mam.d)
        leaf_to_update = mem_to_leaf_index[j]
        c = mam.c
        T_max, j_prime_a, node_pos  = RMaxQST.range_query(T, -1, c-1, len(T_leafs)) 
        # print("C_a:", T_max, T_max + mam.val , j_prime_a, node_pos, leaf_to_update )
        # print("T TREE:", [(s, zz.j, zz.d, zz.Cj, zz.j_max) for s, zz in enumerate(T) if type(zz) != int])
        C_a =  T_max + mam.val #mam.d - mam.c + 1   # add the mam_score to T since disjoint

        if T_max < 0:
            print("BUG", T_max)
            sys.exit()

        
        d = mam.d
        I_max, j_prime_b, node_pos  = RMaxQST.range_query(I, c, c - 1 + overlap_threshold, len(I_leafs))         
        # print( I_max, mam.d, mams[j_prime_b].d, mams[j_prime_b])
        # print("I TREE:", [(s, zz.j, zz.d, zz.Cj, zz.j_max) for s, zz in enumerate(I) if type(zz) != int])
        prev_end = max(mam.c - 1, mams[j_prime_b].d)
        prev_end_diff = mams[j_prime_b].d - (mam.c - 1)
        # (mams[j_prime].d - v.c + 1 ) - 0.0001 if v.x != mams[j_prime].y else v.val - 0.0001)
        ovl_penalty = mams[j_prime_b].d - (mam.c - 1) + 0.0001 if mam.x != mams[j_prime_b].y else 0.0001
        # prev_score = aux_score[j_prime_b]
        # print("prev_score", prev_score)
        # print("C_b:", I_max, I_max  + mam.val - ovl_penalty , j_prime_b, node_pos, leaf_to_update )
        C_b =  I_max  + mam.val - ovl_penalty # (mam.d - prev_end)*mam.val/(mam.d-mam.c + 1) - 0.001 #mam.d #- mams[j_prime_b].d   # add the part of the mam that is not overlapping


        index, value = max_both([C_a, C_b])
        # print("taking:", index, value)
        C[j+1] = value
        # aux_score[j+1] = value
        if index == 0: # Updating with C_a
            j_prime = j_prime_a
        else: # Updating with C_b
            j_prime = j_prime_b


        if j_prime < 0: # any of the additional leaf nodes (with negative index) we add to make number of leafs 2^n
            trace_vector[j+1] = 0
        elif value == 0: # first j (i.e. j=0) 
            trace_vector[j+1]= 0
        else:
            trace_vector[j+1] = j_prime +1

        RMaxQST.update(T, leaf_to_update, value, n) # point update 
        RMaxQST.update(I, leaf_to_update, value, n) # point update 


    solution_index = argmax(C)

    value = C[solution_index]
    solution = []
    while solution_index > 0:
        solution.append(mams[solution_index - 1])  # trace vector is shifted on to the right so need to remove 1 from vectore to get j_index 
        solution_index = trace_vector[solution_index]
    # solutions.append( solution[::-1] )

    non_unique = [x for x in set(C) if C.count(x) > 1]
    unique = True
    if non_unique:
        unique = False
    # sys.exit()
    return tuple(solution[::-1]), value, unique





if __name__ == '__main__':
    mem = namedtuple('Mem', ['x', 'y', 'c', 'd', 'val', 'exon_part_id'])

    mems = [mem(0, 21,1000, 1021, 22, 'id1'), mem(100, 130, 1030,1060, 31, 'id2'), mem(65, 90, 6050, 6075, 26, 'id3'), mem(75, 101, 6060, 6086, 27, 'id4')]
    main(mems)

    import random
    # random.randint(0,100)
    mem_lengths = [ random.randint(1, 10)  for i in range(100)]
    print(mem_lengths)
    positions = [random.randint(0,100) for i in range(100)]
    read_intervals = [(pos, pos + mem_lengths[i] - 1) for i, pos in enumerate(positions)]
    print(read_intervals)

    positions = [random.randint(0,100) for i in range(100)]
    ref_intervals = [(pos, pos + mem_lengths[i] - 1) for i, pos in enumerate(positions)]
    print(ref_intervals)

    mems = [mem(read_iv[0],read_iv[1], ref_intervals[i][0], ref_intervals[i][1], mem_lengths[i], i ) for i, read_iv in enumerate(read_intervals)]
    print(mems)
    # sys.exit()
    main(mems)





import sys
import re
import os
import errno
import itertools

import parasail
import edlib
import dill as pickle 
import collections.abc

# def check_reference_headers(refs):
#     modified = False
#     for header in list(refs.keys()):
#         if header.isdigit() or header == 'X' or header == 'Y':
#             chr_id = 'chr'+ header
#         elif header == 'MT':
#             chr_id = 'chrM'
#         else:
#             chr_id = header

#         # we have modified 
#         if chr_id != header:
#             modified = True
#             seq = refs[header]
#             del refs[header]
#             refs[chr_id] = seq
#     return modified


def update_nested(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_nested(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def remove_read_polyA_ends(seq, threshold_len, to_len):
    end_length_window = min(len(seq)//2, 100)
    seq_list = [ seq[:-end_length_window] ]

    for ch, g in itertools.groupby(seq[-end_length_window:]):
        h_len = sum(1 for x in g)
        # print(ch, h_len, g )
        if h_len > threshold_len and (ch == "A" or ch == "T"):
            seq_list.append(ch*to_len)
        else:
            seq_list.append(ch*h_len)

    seq_mod = "".join([s for s in seq_list])
    return seq_mod

def pickle_dump(filepath, data, filename):
    with open(os.path.join(filepath,filename), 'wb') as f:
        # Pickle the 'data' dictionary using the highest protocol available.
        pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)


def pickle_load(filename):
    with open(filename, 'rb') as f:
        # The protocol version used is detected automatically, so we do not
        # have to specify it.
        data = pickle.load(f)
    return data


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def reverse_complement(string):
    rev_nuc = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'U':'A', 'u':'a', 'a':'t', 'c':'g', 'g':'c', 't':'a', 'N':'N', 'X':'X', 'n':'n', 'Y':'R', 'R':'Y', 'K':'M', 'M':'K', 'S':'S', 'W':'W', 'B':'V', 'V':'B', 'H':'D', 'D':'H', 'y':'r', 'r':'y', 'k':'m', 'm':'k', 's':'s', 'w':'w', 'b':'v', 'v':'b', 'h':'d', 'd':'h'}

    rev_comp = ''.join([rev_nuc[nucl] for nucl in reversed(string)])
    return(rev_comp)


'''
    Below function taken from https://github.com/lh3/readfq/blob/master/readfq.py
'''

def readfq(fp): # this is a generator function
    last = None # this is a buffer keeping the last unprocessed line
    while True: # mimic closure; is it a bad idea?
        if not last: # the first record or a record following a fastq
            for l in fp: # search for the start of the next record
                if l[0] in '>@': # fasta/q header line
                    last = l[:-1] # save this line
                    break
        if not last: break
        name, seqs, last = last[1:].split()[0], [], None
        for l in fp: # read the sequence
            if l[0] in '@+>':
                last = l[:-1]
                break
            seqs.append(l[:-1])
        if not last or last[0] != '+': # this is a fasta record
            yield name, (''.join(seqs), None) # yield a fasta record
            if not last: break
        else: # this is a fastq record
            seq, leng, seqs = ''.join(seqs), 0, []
            for l in fp: # read the quality
                seqs.append(l[:-1])
                leng += len(l) - 1
                if leng >= len(seq): # have read enough quality
                    last = None
                    yield name, (seq, ''.join(seqs)); # yield a fastq record
                    break
            if last: # reach EOF before reading enough quality
                yield name, (seq, None) # yield a fasta record instead
                break



def cigar_to_seq(cigar, query, ref):
    cigar_tuples = []
    result = re.split(r'[=DXSMI]+', cigar)
    cig_pos = 0
    for length in result[:-1]:
        cig_pos += len(length)
        type_ = cigar[cig_pos]
        cig_pos += 1
        cigar_tuples.append((int(length), type_ ))

    r_index = 0
    q_index = 0
    q_aln = []
    r_aln = []
    for length_ , type_ in cigar_tuples:
        if type_ == "=" or type_ == "X":
            q_aln.append(query[q_index : q_index + length_])
            r_aln.append(ref[r_index : r_index + length_])

            r_index += length_
            q_index += length_
        
        elif  type_ == "I":
            # insertion w.r.t. reference
            r_aln.append('-' * length_)
            q_aln.append(query[q_index: q_index + length_])
            #  only query index change
            q_index += length_

        elif type_ == 'D':
            # deletion w.r.t. reference
            r_aln.append(ref[r_index: r_index + length_])
            q_aln.append('-' * length_)
            #  only ref index change
            r_index += length_
        
        else:
            print("error")
            print(cigar)
            sys.exit()

    return  "".join([s for s in q_aln]), "".join([s for s in r_aln]), cigar_tuples


def edlib_alignment(read_seq, ref_seq, aln_mode="NW"):
    result = edlib.align(read_seq, ref_seq, task="path", mode=aln_mode)
    cigar_string = result["cigar"]
    start, stop = result['locations'][0]
    read_alignment, ref_alignment, cigar_tuples = cigar_to_seq(cigar_string, read_seq, ref_seq[start: stop])
    read_alignment = "-"*start + read_alignment + "-"* (len(ref_seq)-stop - 1)
    ref_alignment = ref_seq[:start] + ref_alignment + ref_seq[stop:]
    # cigar_tuples.insert(0, (len(ref_seq[:start]),"D"))
    # cigar_tuples.append((len(ref_seq[stop:]),"D"))
    # print( len(read_alignment), len(ref_alignment))
    assert len(read_alignment) == len(ref_alignment)
    return read_alignment, ref_alignment, result['editDistance']

def mkdir_p(path):
    try:
        os.makedirs(path)
        print("creating", path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

def parasail_alignment(s1, s2, match_score = 2, mismatch_penalty = -2, opening_penalty = 3, gap_ext = 1):
    user_matrix = parasail.matrix_create("ACGT", match_score, mismatch_penalty)
    result = parasail.sg_trace_scan_16(s1, s2, opening_penalty, gap_ext, user_matrix)
    if result.saturated:
        print("SATURATED!",len(s1), len(s2))
        result = parasail.sg_trace_scan_32(s1, s2, opening_penalty, gap_ext, user_matrix)
        print("computed 32 bit instead")

    # difference in how to obtain string from parasail between python v2 and v3... 
    if sys.version_info[0] < 3:
        cigar_string = str(result.cigar.decode).decode('utf-8')
    else:
        cigar_string = str(result.cigar.decode, 'utf-8')
    s1_alignment, s2_alignment, cigar_tuples = cigar_to_seq(cigar_string, s1, s2)
    # print(result.score, len(s1), len(s2))
    # print(s1_alignment)
    # print(s2_alignment)
    # print(cigar_string)
    # sys.exit()
    # print(dir(result))
    # print(result.end_query, result.end_ref, result.len_query, result.len_ref, result.length, result.matches)
    # print()
    return s1_alignment, s2_alignment, cigar_string, cigar_tuples, result.score

    # # Rolling window of matching blocks
    # match_vector = [ 1 if n1 == n2 else 0 for n1, n2 in zip(s1_alignment, s2_alignment) ]    
    # match_window = deque(match_vector[:k]) # initialization
    # current_match_count = sum(match_window)
    # aligned_region = []
    # if current_match_count >= match_id:
    #     aligned_region.append(1)
    # else:
    #     aligned_region.append(0)


    # for new_m_state in match_vector[k:]:
    #     prev_m_state = match_window.popleft()
    #     current_match_count = current_match_count - prev_m_state + new_m_state 
    #     match_window.append(new_m_state)
        
    #     if current_match_count >= match_id:
    #         aligned_region.append(1)
    #     else:        
    #         aligned_region.append(0)

    # # print("".join([str(m) for m in aligned_region]))
    # # print("Aligned ratio (tot aligned/len(seq1):", sum(aligned_region)/float(len(s1)))
    # alignment_ratio = sum(aligned_region)/float(len(s1))
    # return (s1, s2, (s1_alignment, s2_alignment, alignment_ratio))

def ssw_alignment(s1, s2, match_score = 2, mismatch_penalty = -2, opening_penalty = 3, gap_ext = 1):
    user_matrix = parasail.matrix_create("ACGT", match_score, mismatch_penalty)
    result = parasail.ssw(s1, s2, opening_penalty, gap_ext, user_matrix)
    print(result, type(result), dir(result))
    print(dir(result))
    for attr, value in result.__dict__.items():
        print(attr, value)
    # print(result.ref_begin1, result.ref_end1, result.read_begin1, result.read_end1)
    # print()
    return s1_alignment, s2_alignment, cigar_string, cigar_tuples, result.score

def parasail_local(s1, s2, match_score = 2, mismatch_penalty = -2, opening_penalty = 3, gap_ext = 1):
    user_matrix = parasail.matrix_create("ACGT", match_score, mismatch_penalty)
    result = parasail.sw_trace_scan_16(s1, s2, opening_penalty, gap_ext, user_matrix)
    if result.saturated:
        print("SATURATED!",len(s1), len(s2))
        result = parasail.sg_trace_scan_32(s1, s2, opening_penalty, gap_ext, user_matrix)
        print("computed 32 bit instead")

    # difference in how to obtain string from parasail between python v2 and v3... 
    if sys.version_info[0] < 3:
        cigar_string = str(result.cigar.decode).decode('utf-8')
    else:
        cigar_string = str(result.cigar.decode, 'utf-8')
    s1_alignment, s2_alignment, cigar_tuples = cigar_to_seq(cigar_string, s1[result.cigar.beg_query:result.end_query], s2[result.cigar.beg_ref: result.end_ref])
    # print(result.traceback.ref)
    # print(result.traceback.comp)
    # print(result.traceback.query)
    # print(result.score, len(s1), len(s2))
    print("read",s1_alignment)
    print("Rref",s2_alignment)
    print(result.cigar.beg_query,result.end_query)
    print(result.cigar.beg_ref, result.end_ref)
    print(cigar_string)
    # print(result.cigar.seq)

    # sys.exit()
    # print(dir(result))  
    # for attr, value in result.__dict__.items():
    #     print(attr, value)
    # print(result.end_query, result.end_ref, result.len_query, result.len_ref, result.length, result.matches)
    # print()
    return s1_alignment, s2_alignment, cigar_string, cigar_tuples, result.score


def find_all_paths(graph, start, end):
    path  = []
    paths = []
    queue = [(start, end, path)]
    while queue:
        start, end, path = queue.pop()
        # print( 'PATH', path)

        path = path + [start]
        if start == end:
            paths.append(path)
        for node in set(graph[start]).difference(path):
            queue.append((node, end, path))
    return paths



from modules import help_functions

def contains(sub, pri):
    M, N = len(pri), len(sub)
    i, LAST = 0, M-N+1
    while True:
        try:
            found = pri.index(sub[0], i, LAST) # find first elem in sub
        except ValueError:
            return False
        if pri[found:found+N] == sub:
            return True
        else:
            i = found+1

def main(chr_id, predicted_splices, splices_to_transcripts, transcripts_to_splices, all_splice_pairs_annotations, all_splice_sites_annotations):
    # FSM
    transcript = ''
    if len(predicted_splices) == 0:
         return "NO_SPLICE", transcript       

    if  tuple(predicted_splices) in splices_to_transcripts[chr_id]:
        pred_transcripts = splices_to_transcripts[chr_id][tuple(predicted_splices)]
        if type(pred_transcripts) == str:
            transcript = pred_transcripts
        else:
            transcript = ",".join( tr for tr in splices_to_transcripts[chr_id][tuple(predicted_splices)])  
        # print()
        # print('Found, FSM:', transcript, predicted_splices)
        # print()
        return "FSM", transcript

    # NIC 
    all_splice_sites_annotations_chromosome = all_splice_sites_annotations[chr_id] 
    is_nic = True
    for start, stop in predicted_splices:
        if start not in all_splice_sites_annotations_chromosome or stop not in all_splice_sites_annotations_chromosome:
            is_nic = False
    if is_nic:
        all_splice_pairs_annotations_chromosome = all_splice_pairs_annotations[chr_id]
        is_nic_comb = True
        for start, stop in predicted_splices:
            if (start, stop) not in all_splice_pairs_annotations_chromosome:
                is_nic_comb = False


        if is_nic_comb:
            # print()
            # print('Found, NIC/ISM (new combination /or incomplete number of exons):', tuple(predicted_splices) )
            # print()             
            # for ann_tr in splices_to_transcripts[chr_id]:
            #     print(splices_to_transcripts[chr_id][ann_tr] ,ann_tr)

            return  "ISM/NIC_known", transcript

        else:
            # print()
            # print('Found, NIC (new donor-acceptor pair):', tuple(predicted_splices) )
            # print()   
            return   "NIC_novel", transcript   

    # ISM

    # hits = []
    # for splice_pair in predicted_splices:
    #     if splice_pair in all_splice_pairs_annotations[chr_id]:
    #         hits.append(all_splice_pairs_annotations[chr_id][splice_pair])

    # faster looping than above:
    hits = [all_splice_pairs_annotations[chr_id][splice_pair] for splice_pair in predicted_splices if splice_pair in all_splice_pairs_annotations[chr_id]]
    # print("LOOOOL", chr_id, hits)
    # print(predicted_splices)
    if hits:
        in_all_pairs = set.intersection(*hits)
    else:
        in_all_pairs = set()
    # print(in_all_pairs)
    for transcript_id in in_all_pairs:
        transcript_splices = transcripts_to_splices[chr_id][transcript_id]
        # print('LOOOOOL', predicted_splices, transcript_splices)
        if contains(predicted_splices, transcript_splices):
            # print("Found, ISM to", transcript_id )
            transcript = transcript_id
            return "ISM", transcript
        else:
            # print(predicted_splices, transcript)
            pass


    
    # help_functions.eprint()
    # help_functions.eprint('NNC:', tuple(predicted_splices) )
    # help_functions.eprint()      
    # for start, stop in predicted_splices:
    #     if start not in all_splice_sites_annotations_chromosome:
    #         help_functions.eprint(start) 
    #     if stop not in all_splice_sites_annotations_chromosome:
    #         help_functions.eprint(stop) 

    return "NNC", transcript



import pysam
from itertools import groupby
from operator import itemgetter
import re

def get_segments(read_aln, ref_aln, predicted_exons):
    segments = []
    ref_seq_break_points = set()
    prev = 0
    # print(predicted_exons)
    for p1,p2 in predicted_exons:
        ref_seq_break_points.add( p2 - p1 + prev )
        prev += p2-p1
    
    # print('ref_seq_break_points', ref_seq_break_points)

    curr_ref_pos = 0
    ref_aln_break_points = []
    for i, n in enumerate(ref_aln):
        if curr_ref_pos in ref_seq_break_points:
            ref_aln_break_points.append(i)

        if n != '-':
            curr_ref_pos += 1
    # print(curr_ref_pos)
    if n != '-' and curr_ref_pos in ref_seq_break_points: 
        ref_aln_break_points.append(i)


    # remove any consecutive breakpoints caused by insertions in junctions
    ref_aln_break_points_no_consecutive = []
    for k, g in groupby( enumerate(ref_aln_break_points), key=lambda x: x[0] - x[1]):
        consecutive_group_of_coords = list(map(itemgetter(1), g))
        # print(consecutive_group_of_coords)
        ref_aln_break_points_no_consecutive.append(consecutive_group_of_coords[0])

    # print('ref_aln_break_points', ref_aln_break_points)
    # print('ref_aln_break_points_no_consecutive', ref_aln_break_points_no_consecutive)
    # print(len(ref_aln_break_points_no_consecutive))
    e_start = 0
    for i,e_stop in enumerate(ref_aln_break_points_no_consecutive):
        if i == len(ref_aln_break_points_no_consecutive) - 1:
            segments.append( (read_aln[ e_start : ], ref_aln[ e_start : ]) )
        else:
            segments.append( (read_aln[ e_start : e_stop ], ref_aln[ e_start : e_stop ]) )
        e_start = e_stop
    return segments

def get_type(n1, n2):
    if n1 == n2:
        return '='
    elif n1 == '-':
        return 'D'
    elif n2 == '-':
        return 'I'
    else:
        return 'X'

def get_cigars(segments):
    segments_cigar = []
    first, last = 0, len(segments) - 1
    start_offset = 0
    for i, (read, ref) in enumerate(segments):
        # print()
        cigar_types = []
        cigar_lengths = []
        prev_type = get_type(read[0], ref[0])
        length = 1
        for n1,n2 in zip(read[1:], ref[1:]):
            curr_type = get_type(n1, n2)
            if curr_type == prev_type:
                length += 1
            else:
                cigar_types.append(prev_type)
                cigar_lengths.append(length)
                # cigar_tuples.append((length, prev_type))
                length = 1
                prev_type = curr_type
        
        cigar_types.append(prev_type)
        cigar_lengths.append(length)
        # cigar_tuples.append((length, prev_type))

        if i == first:
            if cigar_types[0] == 'D':
                start_offset = cigar_lengths[0]
                del cigar_lengths[0]
                del cigar_types[0]

            elif cigar_types[0] == 'I':
                cigar_types[0] = "S"
        if i == last:
            if cigar_types[-1] == 'D':
                del cigar_lengths[-1]
                del cigar_types[-1]
            elif cigar_types[-1] == 'I':
                cigar_types[-1] = "S"

        segment_cigar = "".join([str(l)+t for l,t in zip(cigar_lengths,cigar_types) ])
        segments_cigar.append(segment_cigar)

    # print(start_offset, 'segments_cigar', segments_cigar)
    return segments_cigar, start_offset #"".join([str(length)+ type_ for length, type_ in c ])


# def modify_beginning(c, p):
#     add_N_beg = 0
#     match_beg = re.match(p, c)
#     if match_beg:
#         # print('heeeej',c)
#         m = match_beg.group()
#         add_N_beg  = int(m[:-1])
#         c = c[len(m):]
#     return c, add_N_beg


# def modify_end(c,p):
#     add_N_end = 0
#     c_rev = c[::-1]
#     match_end = re.match(p, c_rev)
#     if match_end:
#         # print('holllla',c)
#         m_rev = match_end.group()
#         m = m_rev[::-1]
#         add_N_end  = int(m[:-1])
#         c = c[:-len(m)]
#     return c, add_N_end


def get_genomic_cigar(read_aln, ref_aln, predicted_exons):
    # print('here', read_aln)
    # print('here', ref_aln)
    segments = get_segments(read_aln, ref_aln, predicted_exons)
    cigars, start_offset = get_cigars(segments)
    # print('cigar segments', cigars)

    # ######## ORIGINAL  ###########################
    # for c in cigars:
    #     print(c)
    genomic_cigar = []
    intron_lengths = [e2[0] - e1[1] for e1, e2 in zip(predicted_exons[:-1], predicted_exons[1:])]
    # print(intron_lengths)
    for i in range(len(cigars)):
        if i <= len(intron_lengths) -1:
            genomic_cigar.append( cigars[i] + '{0}N'.format( intron_lengths[i] ) )
        else:
            genomic_cigar.append( cigars[i]  )
    # ################################################

 

    # ######################################################################
    # ################ MODIFYING CIGAR AROUND INTRONS  ######################
    # intron_lengths = [e2[0] - e1[1] for e1, e2 in zip(predicted_exons[:-1], predicted_exons[1:])]
    # p = "[1-9]+D"
    # p_rev = "D[1-9]+"
    # if len(cigars) > 1:
    #     # for i, (c1,c2) in enumerate(zip(cigars[:-1], cigars[1:])):
    #     for i, c in enumerate(cigars):
    #         if i == 0: # first
    #             #check only end
    #             c, n_add_end = modify_end(c, p_rev)
    #             intron_lengths[i] += n_add_end

    #         elif i == len(cigars) - 1: # last
    #             #check only beginning
    #             c, n_add_beg = modify_beginning(c, p)
    #             intron_lengths[i-1] += n_add_beg 
                
    #         else: # middle
    #             c, n_add_beg = modify_beginning(c, p)
    #             c, n_add_end = modify_end(c, p_rev)
    #             intron_lengths[i-1] += n_add_beg 
    #             intron_lengths[i] += n_add_end

    #         cigars[i] = c
    # # print()
    # # for c in cigars:
    # #     print(c)
    # genomic_cigar = []
    # # print(intron_lengths)
    # for i in range(len(cigars)):
    #     if i <= len(intron_lengths) -1:
    #         genomic_cigar.append( cigars[i] + '{0}N'.format( intron_lengths[i] ) )
    #     else:
    #         genomic_cigar.append( cigars[i]  )
    # ###########################
    # ###########################

    genomic_cigar = "".join(s for s in genomic_cigar)

    return genomic_cigar, start_offset


def edit_distance(cigar):
    ed = 0
    cigar_tuples = []
    result = re.split(r'[=DXSMIN]+', cigar)
    cig_pos = 0
    for length in result[:-1]:
        cig_pos += len(length)
        type_ = cigar[cig_pos]
        cig_pos += 1
        cigar_tuples.append((int(length), type_ ))

    ed_types = {"X", "I", "D"}
    for length_, type_ in cigar_tuples:
        if type_ in ed_types:
            ed += length_
    return ed 


def main(read_id, read_seq, ref_id, classification, predicted_exons, read_aln, ref_aln, annotated_to_transcript_id, alignment_outfile, is_rc, is_secondary, map_score, aln_score = 0):
    # print(ref_id, classification, predicted_exons, read_aln, ref_aln, alignment_outfile)
    read_sam_entry = pysam.AlignedSegment(alignment_outfile.header)
    if classification != 'unaligned':
        genomic_cigar, start_offset = get_genomic_cigar(read_aln, ref_aln, predicted_exons)
        # if genomic_cigar == "":
        #     print(classification, is_rc, is_secondary, aln_score)
        #     print('genomic cigar:', genomic_cigar, read_id)
        #     print(read_aln)
        #     print(ref_aln)
        #     print(predicted_exons)
        if is_secondary and is_rc:
            read_sam_entry.flag = 256 + 16 
        elif is_secondary:
            read_sam_entry.flag = 256
        elif is_rc:
            read_sam_entry.flag = 16 
        else:
            read_sam_entry.flag = 0 

        read_sam_entry.reference_name = ref_id
        read_sam_entry.mapping_quality = 60 # TODO: calculate mapping quality 

        read_sam_entry.cigarstring = genomic_cigar 
        read_sam_entry.reference_start = predicted_exons[0][0] + start_offset
        read_sam_entry.mapping_quality = map_score 
        # print(predicted_exons[0][0], start_offset)
        read_sam_entry.set_tag('XA', annotated_to_transcript_id)
        read_sam_entry.set_tag('XC', classification)
        read_sam_entry.set_tag('NM', edit_distance(genomic_cigar))
    else:
        read_sam_entry.cigarstring = '*'
        read_sam_entry.reference_start = -1
        read_sam_entry.flag = 4


    read_sam_entry.query_sequence  = read_seq
    read_sam_entry.query_name = read_id




    alignment_outfile.write(read_sam_entry)
    # sys.exit()


